/*
   Copyright (c) JRL CNRS-AIST,
   @author :
                Olivier STASSE
                Alireza NAKHAEI

   All rights reserved.

   Redistribution and use in source and binary forms, with or without 
modification,
   are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright 
notice,
   this list of conditions and the following disclaimer in the 
documentation and/or other materials provided with the distribution.
   * Neither the name of the <ORGANIZATION> nor the names of its 
contributors
   may be used to endorse or promote products derived from this software 
without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS
   OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY
   AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
SHALL THE COPYRIGHT OWNER
   OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
   IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef _WM_GRID_3D_H_
#define _WM_GRID_3D_H_

/* either boost or libimages3d image structure */
#define WM3D_HAS_BOOST

#include <iostream>
#include <string>
#include <fstream>
#include <cmath>

#include <sstream>

#include <stdio.h>      // standard input/output
#include <vector>       // stl vector header

#include <sys/time.h>

#include <stdlib.h>

#define DEFAULT_RESOLUTION	(0.05) // [m]
#define DEFAULT_SENSITIVITY	(0)

#define DEFAULT_MAP_AREA_XSTART	0 // [m]
#define DEFAULT_MAP_AREA_YSTART	-1 // [m]
#define DEFAULT_MAP_AREA_ZSTART	1 // [m]
#define DEFAULT_MAP_AREA_XEND	1 // [m]
#define DEFAULT_MAP_AREA_YEND	-1 // [m]
#define DEFAULT_MAP_AREA_ZEND	1 // [m]
#define UNKNOWN_MIN 0.3
#define UNKNOWN_MAX 3
#define ROBOT_BOX_X 0.75
#define ROBOT_BOX_Y 1.2
#define ROBOT_BOX_Z 1.9


#ifdef WM3D_HAS_BOOST
# include <boost/numeric/ublas/vector.hpp>
# include <boost/numeric/ublas/matrix.hpp>
# include <boost/numeric/ublas/io.hpp>
#endif
#ifdef WM3D_HAS_LIBIMAGES3D
# include "t3d/v3d.h"
# include "t3d/t3d.h"
#ifdef __cplusplus
extern "C" {
#endif
# include "libimages3d.h"
#ifdef __cplusplus
};
#endif

#endif


//
using namespace::std;
// namespace WorldModel3D

//{


	typedef enum {WM_OK, WM_ERROR} wmStatus;

	typedef struct grid
	{
	unsigned occupied:
		7;
		double odd;
                double variance;
                char check;
	}
	grid;


	typedef struct box_area
	{
		double xstart, ystart, zstart;
		double xend, yend,zend;
	}
	box_area;



	class CwmGrid3D
	{
		public:

			//  Constructor
			CwmGrid3D();

			// Destructor
			~CwmGrid3D();


#ifdef WM3D_HAS_BOOST
			
			/**
			   \brief Set Sensor position and orientation according to a 4x4 matrix [m].
			*/
			wmStatus setSensorPosition ( boost::numeric::ublas::matrix<float> inSensorPosition );


			/**
			   \brief Get Sensor position and orientation according to a 4x4 matrix.
			*/
			wmStatus getSensorPosition ( boost::numeric::ublas::matrix<float> outSensorPosition );


#endif /* WM3D_HAS_BOOST */
#ifdef WM3D_HAS_LIBIMAGES3D

			/**
			   \brief Set Sensor position and orientation according to a 4x4 matrix [m].
			*/
			wmStatus setSensorPosition ( const T3D &inSensorPosition );

			/**
			   \brief Get Sensor position and orientation according to a 4x4 matrix.
			*/
			wmStatus getSensorPosition ( T3D &outSensorPosition );

#endif /* WM3D_HAS_LIBIMAGES3D */


			/**
			   \brief Update map regarding the last rangemap which was set.
			*/
			wmStatus updateMap();

			/**
			   \brief Returns the pointer to the data stored at (x,y,z).
			*/
			grid * xyz2grid ( double x, double y, double z );

			/**
			   \brief Returns the position of the grid [i].
			*/
			wmStatus grid2xyz ( int i, double &x, double &y, double &z );

			/**
			   \brief Returns the pointer to the data stored at (x,y,z).
			*/
			grid * xyz2grid ( int x, int y, int z );

			/**
			   \brief Allocate the memory needed for the map.
			*/
			wmStatus allocateMap ( double xstart, double ystart, double zstart,
			                       double xend, double yend, double zend );

			/**
			   \brief Allocate the defult free zone in the environmet (a free zone which is freezed to free in all update).
			*/
			wmStatus setFreeBox ( double xstart, double ystart, double zstart,
			                      double xend, double yend, double zend );

			/**
			   \brief Change the status of a valume of the envirenment to free.
			*/
			wmStatus makeBoxFree ( double xstart, double ystart, double zstart,
			                       double xend, double yend, double zend );

			/**
			   \brief Change the status of a valume of the envirenment to Solid.
			*/
			wmStatus makeBoxSolid ( double xstart, double ystart, double zstart,
			                       double xend, double yend, double zend );

			/**
			   \brief Reset the alocated map to Unknown.
			*/
			wmStatus clearMap();

			/**
			   \brief Set the size of each grid [m].
			*/
			wmStatus setResolution ( double inResolution );

			/**
			   \brief Set the sensetivity of update engin.
			*/
			//if the number that one grid is observed is mor than sensetivity, it will considered as Solid grid.
			wmStatus setSensitivity ( unsigned int inSensitivity );

			/**
			   \brief Returns grid size.
			*/
			double getResolution();

			/**
			   \brief Returns the geometry of map in a vector as (xstart, ystart, zstart, xend, yend, zend).
			*/
			box_area getMapVolume();

			/**
			   \brief Set the information regarding 3D position of the pixel in image.
			*/
#ifdef WM3D_HAS_BOOST
			wmStatus setRangeMap (
			  boost::numeric::ublas::matrix<float> inRangeMap );
#endif
#ifdef WM3D_HAS_LIBIMAGES3D
			wmStatus setRangeMap ( DATA_IM3D *inRangeMap );
#endif

			/**
			   \brief Returns the number of grid along x.
			*/
			int getMapXN();

			/**
			   \brief Returns the number of grid along y.
			*/
			int getMapYN();

			/**
			   \brief Returns the number of grid along z.
			*/
			int getMapZN();

			/**
			   \brief Returns the number of grid.
			*/
			grid * getGridMap();

			/**
			   \brief Export the result of calculation (array of grids) in a txt file for graphic illustration.
			*/
			wmStatus exportOccupancyMap(string fileName);

			/**
			   \brief Export vrml file of the 3D model. by choosing exportMode equal 1, VRML of solid and unknown part are generated seperately.
			*/
			wmStatus exportVRML(string modelFileName , string solidFileName, string unknownFileName,int exportMode);

			/**
			   \brief Import a 3D model.
			*/
                        wmStatus importMap ( string fileName );

			/**
			   \brief Reports all parametere regarding the last update.
			*/
			wmStatus reportCurrentStatus();

			/**
			   \brief Sets camera parameter.
			*/
			wmStatus setCameraParameters ( double inWidthAngle , double inHeightAngle,double inObservableDistance );

			/**
			   \brief Get observable distance of camera.
			*/
			double getObservableDistance ();

			/**
			   \brief Filter the noise in map.
			*/
			wmStatus refreshMap();

			/**
			   \brief Return the maximum number of unknown voxel which can be updated based on the robot config.
			*/
			double getVisibility( double inX, double inY, double inTheta, double goalX, double goalY, double goalTheta );

			/**
			   \brief Set the maximum covering angle in a robot bounding box configuration.
			*/
			wmStatus setCoveringAngle(double inHeightAngleMin, double inHeightAngleMax, double inWidthAngleMin, double inWidthAngleMax);

			/**
			   \brief Set the state of each voxel in map by entering an state vector:0-Free, 1-Occupied, 2-Unknown.
			*/
			wmStatus setGridState( std::vector<char> inStateVector);

			/**
			   \brief Set the robot position.
			*/
			wmStatus setRobotBoxPos( double inX, double inY, double inTheta);

			/**
			   \brief Set the robot bounding box as Free.
			*/
			wmStatus setRobotBoxFree ();

		protected:

			/**
			   \brief Return the odd by getting the distance from a 3D observed point.
			*/
			double getGridParameter ( double inMean ,double inVariance, double inDistance, int k );

			/**
			   \brief Sensor's position.
			*/
#ifdef WM3D_HAS_BOOST
			boost::numeric::ublas::matrix<float> attSensorPosition;
#endif
#ifdef WM3D_HAS_LIBIMAGES3D
			T3D attSensorPosition;
#endif

			/**
			   \brief Number of grid along X, Y and Z.
			*/
			int attMapXN, attMapYN, attMapZN;

			/**
			   \brief Geometric information of Map.
			*/
			box_area attMapArea;

			/**
			   \brief Geometric information of free volume of Map (Initial position of robot).
			*/
			box_area attFreeBox;

			/**
			   \brief The occupancy grid map.
			*/
			grid *attGridMap;

			/**
			   \brief Resolution of the map.
			*/
			double attResolution;

			/**
			   \brief This parametr is the minimum number of point that should be detected in a pixel Volume to be considered as a solid object.
			*/
			unsigned int attSensitivity;

			/**
			   \brief vector that includes the Informatin of 3D position of pixel in image.
			*/
#ifdef WM3D_HAS_BOOST
			boost::numeric::ublas::matrix<float> attRangeMap;
#endif
#ifdef WM3D_HAS_LIBIMAGES3D
			DATA_IM3D *attRangeMap;
#endif

			/**
			   \brief Camera Parametr:the visible angles of camera .
			*/
			double attWidthAngle , attHeightAngle;

			/**
			   \brief the acceptable distance for filtering the valid information of 3D position */
			double attObservableDistance;


			/**
			   \brief Camera Parameters for calculating the visibility.
			*/
                        double attHeightAngleMin, attHeightAngleMax, attWidthAngleMin, attWidthAngleMax;

			/**
			   \brief Robot position in each update.
			*/
                        double attRobotPosX, attRobotPosY, attRobotPosTheta;




	};

// };

#endif
